# -*- coding: utf-8 -*-
"""Version module for WRaINfo."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__version__ = '0.9.5'
__versionalias__ = '2023-07-07_09'



"""Clutter detection module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# imports
# -------

import logging
import xarray as xr
import wradlib as wrl
import numpy as np
from wrainfo.reader import read_single_file
from wrainfo.reader import read_config_file


# fuzzy echo classification based on wRadlib
# -------------------------------------------

def fuzzy_echo_classification(ds, cmap, path, moment="DBZH_sum_bool"):
    """Fuzzy echo classification and clutter identification based on polarimetric variables.

    Parameters
    ----------
    ds : xarray.Dataset
    cmap : xarray.DataArray
        processed clutter map
    path : str
        path to configuration file
    moment: str
        boolean array of the clutter map

    Returns
    -------
    : xarray.Dataset
        xarray.Dataset with a clutter map.
    """
    # extract boolean array of the clutter map
    cmap_bool = cmap[moment]

    trpz_default = {"zdr": [0.7, 1.0, 9999, 9999],
                    "rho": [0.1, 0.15, 9999, 9999],
                    "phi": [15, 20, 10000, 10000],
                    "dop": [-0.2, -0.1, 0.1, 0.2],
                    "map": [1, 1, 9999, 9999],
                    "rho2": [-9999, -9999, 0.95, 0.98],
                    "dr": [-20, -12, 9999, 9999],
                    "cpa": [0.6, 0.9, 9999, 9999], }

    weights = {"zdr": 0.4,
               "rho": 0.4,
               "rho2": 0.4,
               "phi": 0.1,
               "dop": 0.1,
               "map": 0.5}

    dat = dict(rho=ds.RHOHV.values,
               phi=ds.PHIDP.values,
               ref=ds.DBZH.values,
               rho2=ds.RHOHV.values,
               dop=ds.VRAD.values,
               zdr=ds.ZDR.values,
               map=cmap_bool.values)

    dims = read_config_file(path=path, selection="dimensions")

    clmap, nanmask = wrl.clutter.classify_echo_fuzzy(dat,
                                                     weights=weights,
                                                     trpz=trpz_default,
                                                     thresh=0.5)

    ds = ds.assign(dict(FUZZ=(dims, clmap)))

    return ds


# remove clutter
# --------------

def dbzh_no_clutter(ds, path):
    """Set DBZH values, where clutter were identified, to NaN-values.

    Parameters
    ----------
    ds : xarray.Dataset
    path : str
        path to configuration file

    Returns
    -------
    : xarray.Dataset
        xarray.Dataset with removed clutter.
    """
    dims = read_config_file(path=path, selection="dimensions")

    dbzh_no_clutter = ds.DBZH.where(ds.FUZZ == False)   # noqa E712

    dbzh_no_clutter_attrs = {"standard_name": "radar_equivalent_reflectivity_factor_h",
                             "long_name": "Equivalent reflectivity factor H",
                             "unit": "dBZ", }

    ds["DBZH_no_clutter"] = xr.DataArray(dbzh_no_clutter,
                                         dims=dims,
                                         attrs=dbzh_no_clutter_attrs)

    return ds


def create_clutter_map_sequential(files, threshold=0, grp="dataset1", status=False, logger=logging.getLogger()):
    """Create a clutter map sequential in order to not overload the working memory.

    Parameters
    ----------
    files : list
    threshold : integer
        threshold of the reflectivity
    status : bool
        percent of progress

    Returns
    -------
    : xarray.Dataset
        xarray.Dataset with clutter values.
    """
    file_count = 0
    elapsed = 0
    logger.info("Processing " + str(len(files)) + " files sequentially.")
    file = files[0]
    sum_dbzh_values = None

    for file in files:
        try:
            # read furuno files
            data = read_single_file(file=file, grp=grp)

        except Exception:
            # except errors if a file could not read
            logger.warning("Could not load file:' " + file + "'.")
            continue

        # create a data array of reflectivity values
        if sum_dbzh_values is None:
            sum_dbzh_values = np.zeros_like(data.DBZH[0, :, :].values)

        values = data.DBZH[0, :, :].values

        # extract only correct values without NaN values
        good = ~np.isnan(values)
        sum_dbzh_values[good] += data.DBZH[0, :, :].values[good]

        file_count += 1

        new = round(20 * file_count / len(files), 0)

        if status:
            if new > elapsed:
                elapsed = new
                print(str(elapsed * 5) + "% done")

    logger.info("Creating boolean clutter map.")
    sum_dbzh_bool = sum_dbzh_values > threshold

    # set attributes for the new data array
    sum_dbzh_values_attrs = {"standard_name": "radar_equivalent_reflectivity_factor_h",
                             "long_name": "Equivalent reflectivity factor H",
                             "unit": "dBZ", }

    # create dataset
    logger.info("Creating arrays.")
    data["DBZH_sum_values"] = xr.DataArray(sum_dbzh_values,
                                           dims=["azimuth", "range"],
                                           attrs=sum_dbzh_values_attrs)

    data["DBZH_sum_bool"] = xr.DataArray(sum_dbzh_bool,
                                         dims=["azimuth", "range"],
                                         attrs=sum_dbzh_values_attrs)

    # delete all other moments in the dataset except for clutter map
    variables = ['DBZH_sum_bool', 'DBZH_sum_values']
    cmap = data[variables]

    return cmap



"""Compression module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# imports
# -------

import datetime as dt
import pathlib
import tarfile
import os
import glob
from wrainfo.reader import read_config_file
from wrainfo.reader import get_furuno_path


# create tar.gz files
# -------------------

def compress_furuno_raw_data(start_time,
                             end_time,
                             path,
                             delta=dt.timedelta(days=1),
                             delete_files=False):
    """Create daily tar.gz files of FURUNO raw data.

    Parameters
    ----------
    start_time : datetime.datetime
    end_time : datetime.datetime
    path : str
        Path to configuration file
    delta: datetime.delta
        timestamp in which sequence the data will be processed
    delete_files : bool
        if True than files which are packed into tar.gz file were delete

    Returns
    -------
    : files
        daily tar.gz files in output directory.
    """
    date = start_time
    stop_time = end_time
    file_names = []

    outputdir = read_config_file(path=path, selection="output_directory_tar_gz_files")
    radar_location_identifier = read_config_file(path=path, selection="radar_location_identifier")

    while date < stop_time:

        path_day = get_furuno_path(path=path, start_time=date)

        file_names = sorted(glob.glob(os.path.join(path_day, "*")))

        if len(file_names) == 0:

            date += delta

        else:
            file = file_names[1]
            basename = os.path.basename(file)
            year = basename[5:9]
            month = basename[9:11]
            day = basename[11:13]

            outfilename = f"{radar_location_identifier}_{year}{month}{day}.tar.gz"

            path1 = outputdir + "/"

            # output path will create if not exists
            if not os.path.exists(path1):
                os.makedirs(path1, exist_ok=True)
                # overwrite/remove if exist
                f = pathlib.Path(outfilename)
                f.unlink(missing_ok=True)

            # absolute output path
            outfilename = os.path.join(path1, outfilename)
            if not os.path.exists(outputdir):
                os.makedirs(outputdir)
            # overwrite/remove if exist
            f = pathlib.Path(outfilename)
            f.unlink(missing_ok=True)

            with tarfile.open(outfilename, "w:gz") as tar_handle:
                for file in file_names:
                    tar_handle.add(file, arcname=file[35:])

            if delete_files is True:

                for file in file_names:
                    os.remove(file)
                    print("delete file:", file)

            date += delta

    return True


# extract tar.gz files
# ---------------------

def extract_files(path,
                  out_dir):
    """Extract tar.gz - files.

    Parameters
    ---------
    path : path to configuration file
    out_dir : dir where the extracted tar.gz files are saved

    Returns
    -------
    : files
        Extracted tar.gz files in output directory.
    """
    dir_compressed_files = read_config_file(path=path, selection="output_directory_tar_gz_files")

    flist = sorted(glob.glob(os.path.join(dir_compressed_files, "*")))

    for gz_file in flist:
        file = tarfile.open(gz_file)
        file.extractall(out_dir)
        file.close()

    return True



"""Attenuation correction module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# imports
# -------

import xarray as xr
import numpy as np
import wradlib as wrl
from scipy import integrate


# phase processing
# ----------------

def phase_zphi(phi, rng=1000., start_range=0.):
    """Preprocess of PHIDP.

    Parameters
    ----------
    phi : xarray.DataArray
    rng : int
        range to calculate the window to search precipitation bins
    start_range : int
        distance from radar where the processing is starting

    Returns
    -------
    : xarray.Dataset
        xarray.Dataset of preprocessed PHIDP.
    """
    # select measurements after startrange
    phi = phi.where(phi.range >= start_range)

    # get first rangestep
    range_step = np.diff(phi.range)[0]

    # calculate window (need uneven number)
    nprec = int(rng / range_step)
    if (nprec % 2) == 0:
        nprec += 1

    # create binary array
    phib = xr.where(np.isnan(phi), 0, 1)

    # take nprec range bins and calculate sum of valid values
    phib_sum = phib.rolling(range=nprec, center=True).sum(skipna=True)

    # offset in m
    offset = nprec // 2 * range_step
    # offset in idx
    offset_idx = nprec // 2

    # start_range in m (centered)
    start_range = phib_sum.idxmax(dim="range") - offset
    # start_range in idx (centered)
    start_range_idx = phib_sum.argmax(dim="range") - offset_idx

    # stop_range in m (centered)
    stop_range = phib_sum[..., ::-1].idxmax(dim="range") + offset
    # stop_range in idx (centered)
    stop_range_idx = len(phib_sum.range) - (phib_sum[..., ::-1].argmax(dim="range") - offset_idx) - 2

    first0 = phi.where((phi.range >= start_range) & (phi.range <= start_range + rng),
                       drop=True)

    first_min = first0.min(dim='range', skipna=True)
    first_max = first0.max(dim='range', skipna=True)
    first_mean = first0.mean(dim='range', skipna=True)
    first_median = first0.median(dim='range', skipna=True)

    last0 = phi.where((phi.range >= stop_range - rng) & (phi.range <= stop_range),
                      drop=True)

    last_min = last0.min(dim='range', skipna=True)
    last_max = last0.max(dim='range', skipna=True)
    last_mean = last0.mean(dim='range', skipna=True)
    last_median = last0.median(dim='range', skipna=True)

    #     # get min phase values in specified range
    #     first = phi.where((phi.range >= start_range) & (phi.range <= start_range + rng),
    #                        drop=True).min(dim='range', skipna=True)
    #     # get max phase values in specified range
    #     last = phi.where((phi.range >= stop_range - rng) & (phi.range <= stop_range),
    #                        drop=True).max(dim='range', skipna=True)

    return xr.Dataset(dict(phib=phib_sum,
                           offset=offset,
                           offset_idx=offset_idx,
                           start_range=start_range,
                           stop_range=stop_range,
                           first_min=first_min,
                           first_max=first_max,
                           first_mean=first_mean,
                           first_median=first_median,
                           first_idx=start_range_idx,
                           last_min=last_min,
                           last_max=last_max,
                           last_mean=last_mean,
                           last_median=last_median,
                           last_idx=stop_range_idx,
                           ))


def cumulative_trapezoid(da, coord, **kwargs):
    """Cumulative trapezoid integration.

    Parameters
    ---------
    da : xarray.DataArray
        array with data to integrate
    coord : str
        name of coordinate to integrate over

    Returns
    -------
    : xarray.DataArray
        DataArray with integrated values.
    """
    x = da[coord]
    dim = x.dims[0]
    input_core_dim = [dim]
    output_core_dim = [dim]

    # calculate dx
    dx = da[coord].diff(dim).median()
    if coord == "range":
        dx /= 1000.

    kwargs['dx'] = kwargs.get("dx", dx.values)
    kwargs['axis'] = kwargs.get("axis", -1)
    kwargs["initial"] = kwargs.get("initial", 0)

    return xr.apply_ufunc(integrate.cumulative_trapezoid,
                          da,
                          input_core_dims=[input_core_dim],
                          output_core_dims=[output_core_dim],
                          output_dtypes=['float'],
                          dask='parallelized',
                          kwargs=kwargs,
                          dask_gufunc_kwargs=dict(allow_rechunk=True),)


# zphi-method (Testud et al. 2001)
# -------------------------------

def zphi(refl, cphase, alphax=0.28, bx=0.78):
    """Process of PHIDP, KDP and specific attenuation AH_ZPHI.

    Parameter
    ---------
    refl : xarray.DataArray
        array with reflectivity
    cphase : xarray.DataArray
        array of function phase_zphi
    alphax : float
        threshold, default for x-band: 0.28
    bx : float
        threshold

    Returns
    -------
    : xarray.Dataset
        xarray.Dataset with processed variables.
    """
    # calculate delta-phi
    dphi = cphase.last_median - cphase.first_median

    # if negative set dphi zero
    dphi = dphi.where(dphi >= 0).fillna(0)

    # calculate function of dphi
    fdphi = 10 ** (0.1 * bx * dphi * alphax) - 1

    # extract reflectivity
    zhraw = refl.where((refl.range < cphase.stop_range))

    # transform to linear space and fill NaN with zero
    zax = zhraw.pipe(wrl.trafo.idecibel).fillna(0)

    # transform to decibel again
    # zh_cal = zax.pipe(wrl.trafo.decibel)

    # caclulate power
    # equ. 9
    za = zax ** bx

    # fill NaN with zero
    za_zero = za.fillna(0)

    # calculate cumulative integral over range
    iza_x = 0.46 * bx * cumulative_trapezoid(za_zero, coord='range')

    # get maximum over last axis (range) and subtract cumulative integral
    # equ. 10 (r->r2)
    iza = np.max(iza_x, axis=-1) - iza_x

    # divide by function of delta-phi
    iza_fdphi = iza / fdphi

    # TODO: check if this calculates the correct/wanted iza_first
    # equ. 10 (r1->r2)
    # iza_first = iza_fdphi.isel(range=0)
    iza_first = iza_fdphi.isel(range=cphase.first_idx.load())

    # calculate specific attenuation
    # equ. 8
    att = za / (iza_first + iza)
    phi = 2 * cumulative_trapezoid(att / alphax, coord='range')

    # KDP
    kdp = att / alphax

    # encoding properties for output
    enc = dict(zlib=True, complevel=4, chunksizes=(1,) + phi.shape[1:])
    phi.encoding = enc
    phi.name = "PHIDP_RECALC"
    phi_attrs = wrl.io.xarray.moments_mapping["PHIDP"].copy()
    phi_attrs.pop("gamic")
    phi.attrs = phi_attrs

    att.encoding = enc
    pol = refl.long_name[-1:]
    att_name = f"A{pol.upper()}"
    att.name = att_name + "_ZPHI"
    spec_att = dict(units='dB/km',
                    standard_name=f'specific_attenuation_{pol.lower()}',
                    long_name=f'Specific attenuation {pol.upper()}',
                    short_name=att_name
                    )
    att.attrs = spec_att

    kdp.encoding = enc
    kdp.name = "KDP_RECALC"
    kdp_attrs = wrl.io.xarray.moments_mapping["KDP"].copy()
    kdp_attrs.pop("gamic")
    kdp.attrs = kdp_attrs

    return xr.merge([phi.to_dataset(), att.to_dataset(), kdp.to_dataset()])


# process chain for attenuation correction
# ----------------------------------------

def attenuation_correction(ds, moment, dims=["azimuth", "range"]):
    """Attenuation correction with zphi method.

    Parameters
    ----------
    ds : xarray.DataArray
    moment : xarray.DataArray
        DBZH (clutter corrected)
    dims : list
        dimension of the dataset

    Returns
    -------
    : xarray.Dataset
        xarray.Dataset with attenuation corrected refelctivity and recalculated variables.
    """
    # phase processing
    # ----------------

    cphase = phase_zphi(ds.PHIDP.where((ds.RHOHV > 0.8) & (ds[moment] > 10)), rng=1000., start_range=0)

    ds_out = zphi(ds[moment], cphase)

    # attenuation correction
    # ----------------------

    zhcorr = ds[moment] + 0.28 * ds_out.PHIDP_RECALC

    # concat datasets
    # ---------------

    ds["DBZH_CORR"] = xr.DataArray(zhcorr,
                                   dims=dims,
                                   attrs=ds.DBZH.attrs)

    phidp_recalc = ds_out.PHIDP_RECALC
    ds["PHIDP_RECALC"] = xr.DataArray(phidp_recalc,
                                      dims=dims,
                                      attrs=ds_out.PHIDP_RECALC.attrs)

    ah_zphi = ds_out.AH_ZPHI
    ds["AH_ZPHI"] = xr.DataArray(ah_zphi,
                                 dims=dims,
                                 attrs=ds_out.AH_ZPHI.attrs)

    kdp = ds_out.KDP_RECALC
    ds["KDP_RECALC"] = xr.DataArray(kdp,
                                    dims=dims,
                                    attrs=ds_out.KDP_RECALC.attrs)

    return ds



"""Top-level package for WRaINfo."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# package owner
__author__ = """FernLab"""
__email__ = 'fernlab@gfz-potsdam.de'

# import version
from .version import __version__, __versionalias__   # noqa (E402 + F401)

# import subpackages
from wrainfo import attenuation_corr
from wrainfo import clutter
from wrainfo import compression
from wrainfo import error_flist
from wrainfo import geometry
from wrainfo import precipitation
from wrainfo import process_chains
from wrainfo import reader

__all__ = [
    'attenuation_corr',
    'clutter',
    'compression',
    'error_flist',
    'geometry',
    'precipitation',
    'process_chains',
    'reader'
]



"""Geometry module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# imports
# -------

import wradlib as wrl
import numpy as np
import os
import xarray as xr
from wrainfo.reader import read_config_file


# functions to georeferencing and gridding the Furuno data
# --------------------------------------------------------

def get_target_grid(ds, nb_pixels):
    """Get target grid."""
    xgrid = np.linspace(ds.x.min(), ds.x.max(), nb_pixels, dtype=np.float32)
    ygrid = np.linspace(ds.y.min(), ds.y.max(), nb_pixels, dtype=np.float32)
    grid_xy_raw = np.meshgrid(xgrid, ygrid)
    grid_xy_grid = np.dstack((grid_xy_raw[0], grid_xy_raw[1]))
    return xgrid, ygrid, grid_xy_grid


def get_target_coordinates(grid):
    """Get target coordinates."""
    grid_xy = np.stack((grid[..., 0].ravel(), grid[..., 1].ravel()), axis=-1)
    return grid_xy


def get_source_coordinates(ds):
    """Get source coordinates."""
    xy = np.stack((ds.x.values.ravel(), ds.y.values.ravel()), axis=-1)
    return xy


def create_dataarray(ds, data, moment):
    """Create data array."""
    # get moment attrributes
    mom_attrs = ds[moment].attrs
    # remove _Undetect since it's not in the original data
    mom_attrs.pop("_Undetect", None)
    # create moment DataArray
    mom = xr.DataArray(data.astype(np.float32),
                       dims=["y", "x"],
                       attrs=mom_attrs
                       )
    # fix encoding
    mom.encoding = dict(zlib=True, _FillValue=0., least_significant_digit=2)
    return mom


def create_dataset(ds, moments, xgrid, ygrid):
    """Create xarray dataset."""
    # create x,y DataArrays
    x = xr.DataArray(xgrid, dims=["x"], attrs=dict(standard_name="projection_x_coordinate",
                                                   long_name="x coordinate of projection",
                                                   units="m"))
    y = xr.DataArray(ygrid, dims=["y"], attrs=dict(standard_name="projection_y_coordinate",
                                                   long_name="y coordinate of projection",
                                                   units="m"))
    data_vars = dict()
    data_vars.update(moments)
    data_vars.update({"time": ds.reset_coords().time})

    # create Dataset
    ds_out = xr.Dataset(
        data_vars=data_vars,
        coords={"x": x, "y": y},
    )
    # apply CF Conventions
    ds_out.attrs['Conventions'] = 'CF-1.5'
    return ds_out


# georeferencing and gridding of Furuno data
# ------------------------------------------

def furuno_georeferencing(ds, path):
    """Georeference FURUNO Dataset, Grid/Project moments to raster dataset with EPSG Code.

    Parameters
    ----------
    ds : xarray.Dataset
    moments : list
        list of selected data variables from the dataset
    path : str
        path to configuration file

    Returns
    -------
    : xarray.Dataset
        Georeferenced xarray.Dataset.
    """
    # read configuration file for epsg_code and nb_pixels
    epsg_code = read_config_file(path=path, selection="epsg_code")
    nb_pixels = read_config_file(path=path, selection="nb_pixels")
    moments = read_config_file(path=path, selection="moments_in_processed_files")

    # create projection
    proj = wrl.georef.epsg_to_osr(epsg_code)

    # georeference single sweep
    ds = ds.pipe(wrl.georef.georeference_dataset, proj=proj)

    # get source coordinates
    src = get_source_coordinates(ds)

    # create target grid
    xgrid, ygrid, trg_grid = get_target_grid(ds, nb_pixels)

    # get target coordinates
    trg = get_target_coordinates(trg_grid)

    # setup interpolator
    ip_near = wrl.ipol.Nearest(src, trg)

    # calculate moments, create DataArrays
    moment_dict = dict()
    for mom in moments:
        res = ip_near(ds[mom].values.ravel()).reshape((len(ygrid), len(xgrid)))
        moment_dict[mom] = create_dataarray(ds, res, mom)

    # create output dataset
    ds_out = create_dataset(ds, moment_dict, xgrid, ygrid)

    # write crs (projection) to netcdf_file
    ds_out = ds_out.rio.write_crs(epsg_code, grid_mapping_name="transverse_mercator")
    ds_out.attrs["elevation_angle"] = ds.elevation.values[0]

    # add time dimension
    time1 = [1]
    ds_out = ds_out.expand_dims(time=time1)

    return ds_out


# output as netcdf
# ----------------

def furuno_sweep_to_netcdf(ds,
                           path,
                           data_type=None):
    """Create and save georeferenced NetCDF files.

    Parameter
    ---------
    ds : xarray.Dataset
    path: str
        path to configuration file
    data_type : str
        level of processing (Level1, Level2, or Level3)

    Return
    ------
    : NetCDF files
        NetCDF files in output directory.
    """
    # read configuration file
    radar_location_identifier = read_config_file(path=path, selection="radar_location_identifier")
    outdir = read_config_file(path=path, selection="output_path_processed_files")
    moments = read_config_file(path=path, selection="moments_in_processed_files")

    # choose moments, which will save in the netcdf
    ds = ds[moments]

    # case distinction for extraction of elevation angle
    # use time, elevation, and epsg code to create filename
    if hasattr(ds, "elevation"):
        el = ds.elevation.median().values
        t0 = ds.time.values.astype("M8[s]").astype("O")
    else:
        el = ds.elevation_angle
        t0 = ds.time.values.astype("M8[s]").astype("O")[0]

    if data_type is not None:
        outfilename = f"{radar_location_identifier}_{t0:%Y%m%d}_{t0:%H%M%S}UTC_elev_{el}_{data_type}.nc"
    else:
        outfilename = f"{radar_location_identifier}_{t0:%Y%m%d}_{t0:%H%M%S}UTC_elev_{el}.nc"

    # add time dimension
    if len(ds.dims) == 2:  # noqa E712
        time1 = [1]
        ds = ds.expand_dims(time=time1)

    # extract year, month, day from filename
    year = t0.year
    month = f'{t0.month:02d}'
    day = t0.day

    # create output directory
    path1 = outdir + "/" + str(year) + "/" + str(month) + "/" + str(day) + "/" + "elev_" + str(el)

    # absolute output path
    outfilename = os.path.join(path1, outfilename)

    # output path will create if not exists
    if not os.path.exists(path1):
        os.makedirs(path1, exist_ok=True)

    # check whether output file exists
    if not os.path.isfile(outfilename):
        # output file
        print(f"-- output to {outfilename}")
        ds.to_netcdf(outfilename)

    return True



"""Precipitation estimation module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# imports
# -------

import wradlib as wrl
import xarray as xr
from wrainfo.reader import read_config_file


# quantitative precipitation estimation (QPE): R(Z)
# -------------------------------------------------

def qpe_zr(ds,
           moment,
           path,
           a=200,
           b=1.6):
    """Qualitative precipitation estimation with z-R conversion.

    Parameters
    ----------
    ds : xarray.Dataset
    moment : str
        name of the clutter and attenuation corrected data array in the dataset.
    path : str
        path to configuration file
    a : integer
        parameter for the z-R-conversion
       (Marschall Palmer default value: a = 200, German Weather Service = 256)
    b : float
        parameter for the z-R-conversion
       (Marschall Palmer default value: b = 1.6, German Weather Service = 1.42)

    Returns
    -------
    : xarray.Dataset
        xarray.Dataset with a data array of derived precipitation amounts.
    """
    # read setting from configuration file
    dims = read_config_file(path=path, selection="dimensions")
    interval = read_config_file(path=path, selection="scan_interval")

    # precipitation attributes
    prec_attrs = {"standard_name": "precipitation_amount_z_r_relationship",
                  "long_name": "precipitation_amount_z_r_relationship",
                  "unit": "mm", }

    # QPE Z(R)
    rr = ds[moment].pipe(wrl.trafo.idecibel).pipe(wrl.zr.z_to_r, a=a, b=b)
    depth = wrl.trafo.r_to_depth(rr, interval)

    # add to dataset
    ds["PREC_ZR"] = xr.DataArray(depth, dims=dims, attrs=prec_attrs)

    # fix encoding
    enc = dict(zlib=True, complevel=7, chunksizes=(1,) + ds.PREC_ZR.shape)
    ds.PREC_ZR.encoding = enc

    return ds



"""Process chains module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import glob
import xarray as xr
import pathlib
import datetime as dt
import numpy as np
import wrainfo


# clutter map will be calculated monthly every first day of the month
# -------------------------------------------------------------------

def clutter_chain(start_time,
                  path,
                  elevation_angle="dataset1",
                  days=90,
                  threshold=0,
                  status=True,
                  pattern="_000.scnx.gz",
                  data_type="cmap",
                  logger=logging.getLogger()):
    """Create a ground clutter map and save NetCDF-file.

    Parameters
    ----------
    start_time : datetime.datetime
        start time which files will be read
    path : str
        path to configuration file
    elevation_angle : str
        choose dataset of one elevation angle
        (for example: dataset 1 is elevation angle 0.5°)
    days : integer
        Number of days for which a cmap is created
    threshold : float
        this is a threshold for the DBZH it should not be negative
    status : bool
        percentage of the finished clutter map
    pattern  : str
        extension of the scnx files
        (scnx file: elevation angle 0.5° = "_000.scnx")
    data_type : str
        the clutter map (cmap) were created
    logger : file
         writing status messages to a file

    Returns
    -------
    : files
        NetCDF - files in output directory.
    """
    # read output directory for clutter map from configuration file
    clutter_directory = wrainfo.reader.read_config_file(path=path,
                                                        selection="monthly_clutter_directory")
    radar_location_identifier = wrainfo.reader.read_config_file(path=path,
                                                                selection="radar_location_identifier")

    # create the stop time
    # -----------------------

    stop_time = start_time + dt.timedelta(days=days)

    # create a list of all input files for the defined time period
    # ---------------------------------------------------------------

    flist = wrainfo.reader.create_filelist(path_to_config_file=path,
                                           starttime=start_time,
                                           endtime=stop_time,
                                           pattern=pattern)

    # create clutter map
    # ---------------------
    logger.info("Creating cluttermap.")
    cmap = wrainfo.clutter.create_clutter_map_sequential(files=flist,
                                                         threshold=threshold,
                                                         grp=elevation_angle,
                                                         status=status,
                                                         logger=logger.getChild("wr_furuno.clutter"
                                                                                ".create_clutter_map_sequential"))

    # elevation angle
    # ------------------
    el = cmap.elevation.values[0]

    # create outfilename
    # ---------------------
    outfilename = f"{radar_location_identifier}_{start_time:%Y%m%d}_{stop_time:%Y%m%d}_elev_{el:0.1f}_{data_type}.nc"
    year = str(stop_time.year)
    path1 = clutter_directory + "/" + year

    if not os.path.exists(path1):
        os.makedirs(path1, exist_ok=True)
        # overwrite/remove if exist
        f = pathlib.Path(outfilename)
        f.unlink(missing_ok=True)

    outfilename = os.path.join(path1, outfilename)

    # output file as netcdf
    # ------------------------
    logger.info("Creating cluttermap netcdf output.")
    print(f"-- output to {outfilename}")
    cmap.to_netcdf(outfilename)
    logger.info("Finished cluttermap chain.")

    return True


# static clutter map will be processed each month from 3 last cluttermaps
# -----------------------------------------------------------------------

def static_cmap(path,
                pattern="_elev_0.5_cmap.nc"):
    """Create a static ground clutter map and save NetCDF-file.

    Parameters
    ----------
    path : str
        path to configuration file
    pattern : str
        extension of the netcdf file
        (cmaps for different elevation angles)

    Returns
    -------
    : files
        NetCDF - files in output directory.
    """
    # read path of clutter directory from configuration file
    path_cmap = wrainfo.reader.read_config_file(path=path,
                                                selection="monthly_clutter_directory")
    path_cmap = path_cmap + "*"

    # read radar location identifier from configuration file
    radar_location_identifier = wrainfo.reader.read_config_file(path=path,
                                                                selection="radar_location_identifier")

    # read output directory from configuration file
    outdir = wrainfo.reader.read_config_file(path=path,
                                             selection="static_clutter_directory")

    # load the last 3 cmaps
    cmaps = sorted(glob.glob(os.path.join(path_cmap, "*")))

    # filter for elevation angle
    cmaps_filter = []

    for cmap in cmaps:
        if cmap.endswith(pattern):
            cmaps_filter.append(cmap)
            cmaps_filter = cmaps_filter[-3:]

    # read the last 3 cmaps
    cmap1 = xr.open_dataset(cmaps_filter[0])
    cmap2 = xr.open_dataset(cmaps_filter[1])
    cmap3 = xr.open_dataset(cmaps_filter[2])

    # add cmaps
    cmap_final = cmap1 + cmap2 + cmap3

    # outfilename
    base_name = os.path.basename(cmaps_filter[1])
    el = base_name[28:36]

    outfilename = f"{radar_location_identifier}_static_cluttermap_{el}.nc"

    # output directory
    if not os.path.exists(outdir):
        os.makedirs(outdir, exist_ok=True)
        # overwrite/remove if exist
        f = pathlib.Path(outfilename)
        f.unlink(missing_ok=True)

    outfilename = os.path.join(outdir, outfilename)

    print(f"-- output to {outfilename}")
    cmap_final.to_netcdf(outfilename)

    return True


# weather radar routine
# ---------------------

def wr_routine_furuno(starttime,
                      endtime,
                      delta,
                      path,
                      data_type="Level2a",
                      elevation_angle="dataset1",
                      pattern="_000.scnx.gz",
                      logger=logging.getLogger()):
    """Process of Furuno raw data.

    Parameters
    ----------
    starttime : datetime.datetime
    endtime : datetime.datetime
    delta : datetime.delta
    path : str
        path to configuration file
    data_type : str
        Level of processing
    pattern : extension of the scnx file
             (scnx file: elevation angle 0.5° = "_000.scnx")
    elevation_angle : str
        group of hdf5 dataset
    logger : file
         writing status messages to a file

    Returns
    -------
    : files
        NetCDF - files in output directory
    """
    date = starttime
    stop_time = endtime

    # read settings from configuration file
    re_idx = wrainfo.reader.read_config_file(path=path,
                                             selection="re_index_parameters")

    re_index_parameters = np.arange(re_idx[0], re_idx[1], re_idx[2])

    dims = wrainfo.reader.read_config_file(path=path,
                                           selection="dimensions")

    while date < stop_time:

        # read static cmap
        # ----------------

        logger.info("Read static clutter map")
        stat_cmap = wrainfo.reader.read_config_file(path=path,
                                                    selection="static_cmap")
        cmap = xr.open_dataset(stat_cmap)
        cmap_bool = cmap.DBZH_sum_bool

        # load files
        # ----------
        logger.info("Create list of files to be processed")
        flist_raw = wrainfo.reader.create_filelist(path_to_config_file=path,
                                                   starttime=date,
                                                   endtime=date + delta,
                                                   pattern=pattern)

        # start processing
        # ----------------
        for file in flist_raw:

            # extension of the file
            # ---------------------
            extension = os.path.splitext(file)[1]

            if extension == ".gz":

                # read Furuno scnx data
                # ---------------------
                logger.info("Read Furuno scnx file")
                try:
                    ds = xr.open_dataset(file,
                                         engine="furuno",
                                         group=1,
                                         backend_kwargs=dict(reindex_angle=1.0))

                except Exception:
                    # except errors if a file could not read
                    logger.warning("Could not load scnx file:' " + file + "'.")
                    continue

                # rename VRADH to VRAD
                ds['VRAD'] = ds['VRADH']
                ds = ds.drop(['VRADH'])

            if extension == ".h5":

                # read Furuno hdf5 data
                # ---------------------
                logger.info("Read Furuno hdf5 file")
                try:
                    ds = xr.open_dataset(file,
                                         engine="odim",
                                         group=elevation_angle,
                                         backend_kwargs=dict(reindex_angle=1.0))

                except Exception:
                    # except errors if a file could not read
                    logger.warning("Could not load hdf5 file:' " + file + "'.")
                    continue

            # reindexing azimuth resolution
            # -----------------------------
            ds = ds.reindex(azimuth=re_index_parameters,
                            method="nearest")

            # fix errors by elevation angle
            # -----------------------------
            elevation_angle_ds = ds.elevation.median(dim="azimuth")
            ds["elevation"][:] = elevation_angle_ds

            # SNR filter by RHOHV
            # --------------------

            ds = ds.where(ds.RHOHV > 0.9)

            # remove static clutter with cmap
            # --------------------------------

            dbzh = ds.DBZH.where((cmap_bool == False))   # noqa E712

            dbzh_attrs = ds.DBZH.attrs
            ds["DBZH"] = xr.DataArray(dbzh,
                                      dims=dims,
                                      attrs=dbzh_attrs)

            #  clutter detection
            # --------------------
            logger.info("Clutter classification.")
            ds = wrainfo.clutter.fuzzy_echo_classification(ds,
                                                           cmap=cmap,
                                                           path=path,
                                                           moment="DBZH_sum_bool")

            # set clutter pixel to NA
            ds = wrainfo.clutter.dbzh_no_clutter(ds, path=path)

            # phase processing & attenuation correction (ZPHI method)
            # -------------------------------------------------------
            logger.info("Phase processing and attenuation correction.")
            ds = wrainfo.attenuation_corr.attenuation_correction(ds,
                                                                 moment="DBZH_no_clutter",
                                                                 dims=dims)

            # rain rate retreival
            # --------------------
            logger.info("Create precipitation products.")

            # QPE Z(R)
            logger.info("Create precipitation product with Z(R) relationship.")
            ds = wrainfo.precipitation.qpe_zr(ds,
                                              "DBZH_CORR",
                                              path=path,
                                              a=200,
                                              b=1.6)

            # smoothing
            ds = ds.pad(azimuth=(3, 3),
                        mode="wrap").rolling(azimuth=7,
                                             min_periods=3,
                                             center=True).mean().isel(azimuth=slice(3, -3))

            # georeferencing and gridding
            # ---------------------------
            logger.info("Georeferencing and gridding")

            swp_georef = wrainfo.geometry.furuno_georeferencing(ds,
                                                                path=path)

            wrainfo.geometry.furuno_sweep_to_netcdf(swp_georef,
                                                    path=path,
                                                    data_type=data_type)
            del ds
            del swp_georef

        date += delta

    return True



"""Reader module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# imports
# -------

import datetime as dt
import os
import glob
import re
import xarray as xr
import numpy as np
import pickle
import json
from pathlib import Path


# read configuration file
# -----------------------

def read_config_file(path=".../example_settings_path_dependencies_wr_furuno.json", selection=None):
    """Read wr_furuno config file in json format.

    Parameter
    ---------
    path : str
        Path to config file.
    selection : str
        select setting parameter

    Return
    ------
    : dict
        configured directories in the configuration file.
    """
    path = path.replace("~", str(Path.home()))

    if not Path(path).is_file():
        raise Exception("wrainfo.reader.read_config_file(): Config file'", path, "' not found.")

    file = open(path)
    conf = json.load(file)
    file.close()

    if selection is not None:
        if selection in conf.keys():
            return conf[selection]
        else:
            raise Exception("wrainfo.reader.read_config_file(): Setting parameter'",
                            selection,
                            "' not in settings file '",
                            path,
                            "'.")
    else:
        return conf


# Functions to read the Furuno data
# ----------------------------------

def get_furuno_path(path, start_time=dt.datetime.today()):
    """Create path of Furuno radar data files.

    Parameters
    ----------
    path : str
        Path to configuration file
    start_time : datetime.datetime
        datetime object to select correct folder

    Returns
    -------
    radar_path : str
        Path to radar data
    """
    raw_path = read_config_file(path=path, selection="raw_data_directory")
    subfolder_struct_raw_path = read_config_file(path=path, selection="subfolder_structure_raw_data")

    radar_path = os.path.join(raw_path, subfolder_struct_raw_path)
    return radar_path.format(start_time.year, start_time.month, start_time.day)


def get_file_date_regex(filename):
    """Get regex from filename."""
    # regex for ""%Y-%m-%d--%H:%M:%S"
    reg0 = r"\d{4}.\d{2}.\d{2}..\d{2}.\d{2}.\d{2}"
    # regex for "%Y%m%d%H%M%S"
    reg1 = r"\d{14}"
    # regex for 20220216_085000
    reg2 = r"\d{4}\d{2}\d{2}.\d{2}\d{2}\d{2}"
    for reg in [reg0, reg1, reg2]:
        match = re.search(reg, os.path.basename(filename))
        if match is not None:
            return reg
    return None


def get_datetime_from_filename(filename, regex):
    """Get datetime from filename."""
    fmt = "%Y%m%d%H%M%S"
    match = re.search(regex, os.path.basename(filename))
    match = "".join(re.findall(r"[0-9]+", match.group()))
    return dt.datetime.strptime(match, fmt)


def load_error_flist(path):
    """Load error file list.

    Parameter
    ---------
    path : str
        Path to configuration file

    Returns
    ------
    : list
        list of error files.
    """
    path_error_flist = read_config_file(path=path, selection="error_flist_directory")

    with open(path_error_flist, 'rb') as fp:
        error_files = pickle.load(fp)

    return error_files


def create_filelist(starttime, endtime, path_to_config_file, pattern="_000.scnx.gz"):
    """Create filelist from path_glob and filename dates.

    Parameters
    ----------
    starttime : dt.datetime
        start time
    endtime : dt.datetime
        end time
    path_to_config_file : str
        Path to configuration file
    pattern : str
        extension of the scnx/netcdf file
        (scnx file: elevation angle 0.5° = "_000.scnx.gz")

    Returns
    -------
    : list
        list of files
    """
    flist = []
    flist_scnx = []
    flist_h5 = []
    datetime_list = []
    date = starttime

    path_error_flist = read_config_file(path=path_to_config_file, selection="error_flist_directory")

    if os.path.isfile(path_error_flist):
        error_flist = load_error_flist(path=path_to_config_file)
    else:
        error_flist = []

    while date <= endtime:

        raw_path = get_furuno_path(path=path_to_config_file, start_time=date)

        file_names = sorted(glob.glob(os.path.join(raw_path, "*")))

        # fixed empty list of file_names
        if len(file_names) > 0:
            regex = get_file_date_regex(file_names[0])

            for fname in file_names:
                time = get_datetime_from_filename(fname, regex)
                if time >= date:
                    if time < endtime:
                        if fname.endswith(pattern):
                            if fname not in error_flist:
                                flist_scnx.append(fname)
                        if fname.endswith(".h5"):
                            if fname not in error_flist:
                                flist_h5.append(fname)

            if len(flist_scnx) == 0:
                if len(flist_h5) > 0:
                    for file in flist_h5:
                        flist.append(file)

            if len(flist_scnx) > 0:
                for file in flist_scnx:
                    filetime_scnx = get_datetime_from_filename(file, regex)
                    datetime_list.append(filetime_scnx)
                    flist.append(file)
                if len(flist_h5) > 0:
                    for file in flist_h5:
                        filetime_h5 = get_datetime_from_filename(file, regex)
                        if filetime_h5 not in datetime_list:
                            flist.append(file)

        date = date + dt.timedelta(1)

        flist = sorted(set(flist))

    return flist


# read single file of Furuno data
# (important for create the clutter map)
# ---------------------------------------

def read_single_file(file, grp="dataset1"):
    """Read a single file, reindexes and returns a dataset object.

    Parameters
    ----------
    file : str
        path to file as string
    grp : str
        hdf5 group (elevation angle)

    Returns
    -------
    : xarray Dataset
        xarray Dataset with variables.
    """
    extension = os.path.splitext(file)[1]

    if extension == ".gz":
        # the elevation angle the elevation angle is already selected in the flist!
        tmp = xr.open_mfdataset((file,), engine="wradlib-furuno", group=1,
                                concat_dim="time", combine="nested",
                                backend_kwargs=dict(reindex_angle=1.0))
        # rename VRADH to VRAD
        tmp['VRAD'] = tmp['VRADH']
        tmp = tmp.drop(['VRADH'])

        # rename WRADH to WRAD
        tmp['WRAD'] = tmp['WRADH']
        tmp = tmp.drop(['WRADH'])

    else:
        tmp = xr.open_mfdataset((file,), engine="wradlib-odim", group=grp,
                                concat_dim="time", combine="nested",
                                backend_kwargs=dict(reindex_angle=1.0))

    # TODO add case sensitive decision on these parameters
    re_index_parameters = np.arange(0.25, 360, 0.5)

    ri = tmp.reindex(azimuth=re_index_parameters, method="nearest")
    return ri


# read multiple files
# -------------------

def read_multiple_scnx_files(starttime, endtime, path, pattern):
    """Read multiple files, reindexes and returns a dataset object.

    Parameters
    ----------
    starttime : dt.datetime
        start time
    endtime : dt.datetime
        end time
    path : str
        path to configuration file
    pattern : str
        extension of the scnx/netcdf file
        (scnx file: elevation angle 0.5° = "_000.scnx.gz")

    Returns
    -------
    : xarray Dataset
        xarray Dataset with variables.
    """
    flist = create_filelist(starttime=starttime,
                            endtime=endtime,
                            path_to_config_file=path,
                            pattern=pattern)

    scnx = xr.open_mfdataset(flist,
                             engine="wradlib-furuno",
                             group=1,
                             combine="nested",
                             concat_dim="time",
                             backend_kwargs=dict(reindex_angle=1.0))

    # rename VRADH to VRAD
    scnx['VRAD'] = scnx['VRADH']
    scnx = scnx.drop(['VRADH'])

    # rename WRADH to WRAD
    scnx['WRAD'] = scnx['WRADH']
    scnx = scnx.drop(['WRADH'])

    return scnx


# read processed cluttermap from Furuno
# -------------------------------------

def get_cmap(starttime,
             endtime,
             path,
             elev="0.5",
             timestr="%Y%m%d"):
    """Read processed cluttermap from Furuno.

    Parameters
    ---------
    starttime : datetime.datetime
        datetime - object to select correct files for list
    endtime : datetime.datetime
        datetime - object to select correct files for list
    path : str
        Path to configuration file
    elev : str
        selected elevation angle of cmap

    Returns
    -------
    : xarray.Dataset
        clutter map as xarray.Dataset.
    """
    # read settings from configuration file
    clutter_dir = read_config_file(path=path, selection="monthly_clutter_directory")
    subfolderstruct = read_config_file(path=path, selection="subfolder_structure_clutter_directory")

    # create file structure
    file_struct = os.path.join(f"{clutter_dir}", f"{subfolderstruct}", "*")

    path = file_struct.format(year=starttime.year)

    # list all cmaps in the folder
    cmaps = [cmap for cmap in sorted(glob.glob(path)) if cmap.endswith(".nc")]

    selected_map = None
    last_date = dt.datetime(1900, 1, 1)

    for cmap in cmaps:
        # choose the stop time from filename
        filetime_end = dt.datetime.strptime((os.path.basename(cmap)[19:27]), timestr)
        elev_cmap = os.path.basename(cmap)[33:36]

        if filetime_end > (endtime - dt.timedelta(45)):
            if filetime_end > last_date:
                if elev_cmap == elev:   # noqa E712
                    selected_map = cmap
                    last_date = filetime_end

        if selected_map is None:
            print("wrainfo.reader.get_cmap(): No cmap found for these dates.")

    # read cmap
    cmap = xr.open_dataset(selected_map)

    return cmap



#!/usr/bin/env python

"""The setup script."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

version = {}

with open("wrainfo/version.py") as version_file:
    exec(version_file.read(), version)

req = ['numpy>=1.22.1',
       'pytest>=6.2.5',
       'scipy>=1.7.0',
       'wradlib>=1.15.0',
       'xarray[complete]==0.20.2',
       'rioxarray==0.12.0',
       'GDAL==3.6.*']

req_setup = ['pytest-runner']

req_test = ['pytest>=3', 'pytest-cov', 'pytest-reporter-html1', 'urlchecker==0.0.32']

req_doc = [
    'sphinx>=4.1.1',
    'sphinx-argparse',
    'sphinx-autodoc-typehints',
    'sphinx_rtd_theme',
    'pandoc',
    'nbsphinx']

req_lint = ['flake8', 'pycodestyle', 'pydocstyle']

req_dev = ['twine'] + req_setup + req_test + req_doc + req_lint

setup(
    author="FernLab",
    author_email='fernlab@gfz-potsdam.de',
    python_requires='>=3.7',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
    description="Is a software to process FURUNO weather radar data.",
    entry_points={
        'console_scripts': [
            'wrainfo=bin.wrainfo_cli:main',
        ],
    },
    extras_require={
        "doc": req_doc,
        "test": req_test,
        "lint": req_lint,
        "dev": req_dev
    },
    install_requires=req,
    license="Apache Software License 2.0",
    include_package_data=True,
    keywords=['wrainfo', 'weather radar', 'X-Band', 'preprocessing', 'clutter detection', 'attenuation correction', 'precipitation estimation'],
    long_description=readme,
    name='wrainfo',
    packages=find_packages(include=['wrainfo', 'wrainfo.*']),
    project_urls={
        "Source code": "https://git.gfz-potsdam.de/fernlab/products/furuno/wrainfo",
        "Documentation": "https://fernlab.git-pages.gfz-potsdam.de/products/furuno/wrainfo/doc/",
        "Change log": "https://git.gfz-potsdam.de/fernlab/products/furuno/wrainfo/-/blob/main/HISTORY.rst"
    },
    setup_requires=req_setup,
    test_suite='tests',
    tests_require=req + req_test,
    url='https://git.gfz-potsdam.de/fernlab/products/furuno/wrainfo',
    version=version['__version__'],
    zip_safe=False,
)



"""Create error file list module."""

# WRaINfo, Is a software to process FURUNO weather radar data.
#
# Copyright (c) 2023, FernLab (GFZ Potsdam, fernlab@gfz-potsdam.de)
#
# This software was developed within the context of the RaINfo ("Potential use of
# high resolution weather data in agriculture") project of FernLab funded by
# the Impulse and Networking Fund of the Helmholtz Association.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# imports
# -------

import pickle
from wrainfo.reader import load_error_flist
from wrainfo.reader import create_filelist
from wrainfo.reader import read_config_file


# create error flist
# ------------------

def create_error_filelist(out_dir):
    """Create error file list.

    Parameter
    ---------
    out_dir : str
        Output directory with filename

    Return
    ------
    : file
        empty file in output directory.
    """
    error_flist = []

    with open(out_dir, 'wb') as fh:
        pickle.dump(error_flist, fh)

    print(f"-- output to {out_dir}")

    return True


# update error file list
# ----------------------

def update_error_flist(start_time,
                       end_time,
                       path,
                       pattern='_000.scnx.gz'):
    """Update error file list manually.

    Parameter
    ---------
    start_time : datetime.datetime
    end_time : datetime.datetime
    path : str
        Path to the configuration file
    pattern : str
        extension of the scnx file (elevation angle 0.5° = "_000.scnx.gz")

    Return
    ------
    : list
        list of error files.
    """
    error_flist = load_error_flist(path=path)

    flist = create_filelist(starttime=start_time,
                            endtime=end_time,
                            path=path,
                            pattern=pattern)

    outdir = read_config_file(path=path, selection="output_path_error_flist")

    for fname in flist:
        print("Would you like add this file to error flist?" + "\n" + fname)
        user_input = input('Confirm? [Y/N] ')

        if user_input.lower() in ('y', 'yes'):
            error_flist.append(fname)
            print(fname + "\n" + "successfully added to error file list.")
        elif user_input.lower() in ('n', 'no'):
            continue

    print("Would you like save the changes to output directory?")
    user_input = input('Confirm? [Y/N] ')

    if user_input.lower() in ('y', 'yes'):
        with open(outdir, 'wb') as fp:
            pickle.dump(error_flist, fp)
            print("output to" + "\n" + outdir)
            return True

    elif user_input.lower() in ('n', 'no'):
        return error_flist
